#
# Copyright 2018 Fairphone B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Android device basic representation and management."""

from contextlib import contextmanager
import time
from typing import Any, List

from adb.adb_commands import AdbCommands
from adb.adb_protocol import AuthSigner
import adb.usb_exceptions
import usb1

from .app import App
from . import utils


class AppInstallationError(Exception):
    """An error related to an app installation on a device."""

    def __init__(
        self, device: "Device", app: App, reason: str, error_message: str = None
    ):
        """Initialise the exception.

        Attributes:
            device: The target device.
            app: The app which installation is problematic.
            reason: The high-level reason of the error.
            error_message (optional): The tooling error message.

        """
        self.device = device
        self.app = app
        self.reason = reason
        self.error_message = error_message
        message = "The app `{}` installation failed on {}: {}".format(
            str(self.app), str(self.device), self.reason
        )
        super().__init__(message)


class Device:
    """A device accessible through ADB.

    Well-known and commonly used system properties are available as attributes.
    Some of them really are only shortcuts to ADB commands (and thus requires
    ADB to return a value) while others are cached at initialisation since they
    should not change - even if a new software build is installed.

    Any command that requires an ADB connection, or more generally any call to
    the ADB context manager (`#adb()`), can encounter exceptions linked to the
    USB channel itself (e.g. the device is busy or not there) and to the ADB
    protocol (e.g. the device does not authorise the host). Refer to the ADB
    context manager method for the exact list of possible errors.

    The attrbiutes that requires an active ADB connection will *not* raise any
    exception if something goes wrong. Rather, they will return `None`. On the
    other hand, the other shortcut methods will raise any underlying exception.

    As a matter of conveniency, the ADB context manager is reentrant. As such,
    it can be used to initially detect the availability of a device's ADB
    connection:

        with device.adb() as _adb:
            # Query any attribute you like and interact with the device through
            # ADB all you like.
            model = device.model
            uptime = _adb.Shell("uptime")
            print(f"The {model} has been up for {uptime}")
        except adb.usb_exceptions.DeviceAuthError:
            # The device has not authorised the current host
        except adb.usb_exceptions.DeviceNotFoundError:
            # The device has not beed seen on any USB port
        except usb1.USBError:
            # The USB channel to the device is busy

    An additional effect to be noted is that the physical device *does not need
    to stay connected* to the host once the context manager has been closed.
    This can be particularly useful for long-lived programs that do not need to
    interact with devices for an extended period of time, they will react to a
    device loss lazily (i.e. only when try to address it).

    Attributes:
        build_incremental: The build version incremental "number". Requires ADB.
        fingerprint: The build fingerprint. Requires ADB.
        model: The device model name. Requires ADB.
        product: The device product name. Requires ADB.
        sdk: The build Android SDK version. Requires ADB.
        serial: The device serial number. Cached at initialisation.

    """

    def __init__(self, serial: str, adb_signer: AuthSigner = None):
        """Initialise the device.

        Attributes:
            serial: The device serial number.
            adb_signer (optional): The `AuthSigner`, or a new `AuthSigner` from
                `utils.auth_signer_from_user_keys()`.

        """
        self._adb_properties = {}
        if not adb_signer:
            adb_signer = utils.auth_signer_from_user_keys()
        self.adb_config = {"serial": serial, "rsa_keys": [adb_signer]}

        # Cache immutable properties
        self.serial = serial
        # Provide shortcuts to system properties
        self._register_adb_property(
            "build_incremental", "ro.build.version.incremental"
        )
        self._register_adb_property("fingerprint", "ro.build.fingerprint")
        self._register_adb_property("model", "ro.product.model")
        self._register_adb_property("product", "ro.product.name")
        self._register_adb_property("sdk", "ro.build.version.sdk", int)

        # Internal variables
        self.__adb = AdbCommands()

    def __str__(self) -> str:
        """Return a simple representation of the device."""
        return f"{self.product}<{self.serial}>"

    def __repr__(self) -> str:
        """Return an unambiguous representation of the device."""
        return f"{self.product}<{self.serial}> ({self.fingerprint})"

    def __getattr__(self, item) -> Any:
        """Retrieve ADB system properties.

        Attributes:
            item: The cached property name.
        Returns:
            The value of the property, or `None` if the device is not reachable.
        Raises:
            AttributeError: If the system property is not known.

        """
        if item in self._adb_properties.keys():
            try:
                return self._adb_properties[item]["type"](
                    self.getprop(self._adb_properties[item]["key"])
                )
            except (
                adb.usb_exceptions.DeviceAuthError,
                adb.usb_exceptions.DeviceNotFoundError,
                usb1.USBError,
            ):
                return None
        raise AttributeError(f"Unknown system property `{item}`")

    def _register_adb_property(
        self, name: str, key: str, value_type: type = str
    ) -> None:
        """Add a system property to the system properties mapping.

        Arguments:
            name: The attribute name that will be mapped to the property.
            key: The system property key.
            value_type: The type the value should be converted to when worked
                with.

        """
        self._adb_properties[name] = {"key": key, "type": value_type}

    @property
    def storage_dir(self) -> str:
        """Return the root storage directory."""
        return "/sdcard"

    @classmethod
    def get_all(cls, adb_signer: AuthSigner = None) -> List["Device"]:
        """Create a list of all devices reachable by the host.

        TODO: They must be authorised.

        Attributes:
            adb_signer (optional): The `AuthSigner`, or a new `AuthSigner` from
                `utils.auth_signer_from_user_keys()`.
        Returns:
            The list of new device objects.

        """
        if not adb_signer:
            adb_signer = utils.auth_signer_from_user_keys()
        return [
            Device(serial=serial, adb_signer=adb_signer)
            for serial in cls.list_serials()
        ]

    @classmethod
    def list_serials(cls) -> List[str]:
        """List all device serial numbers reachable by the host.

        Returns:
            The list of serial numbers belonging to the reachable devices.

        """
        return [handle.serial_number for handle in AdbCommands.Devices()]

    @contextmanager
    def adb(self):
        """Create an ADB context manager for the device.

        Connect to the device before yielding the resulting `AdbCommands`
        object. At the end of the context, close the connection to free the USB
        channel.

        The context manager is considered reentrant since the `AdbCommands`
        instance will be re-used if it already exists, that is if the USB
        channel is already kept busy. In a case of reentrance, exiting the
        context would not close the connection, only the initiating context will
        close it.

        Returns:
            The `AdbCommands` instance bound to this device.
        Raises:
            adb.usb_exceptions.DeviceAuthError: If the keys passed to the
                constructor (the `AuthSigner` object) are not authorised on the
                device itself.
            adb.usb_exceptions.DeviceNotFoundError: If the device is not found.
            usb1.USBError: If the device is busy (i.e. already connected).

        """
        if self.__adb.GetState():
            yield self.__adb
        else:
            try:
                self.__adb.ConnectDevice(**self.adb_config)
                yield self.__adb
            finally:
                self.__adb.Close()

    def getprop(self, key: str) -> str:
        """Get a system property of the device.

        Propagate any exception raised by `#adb()`.

        Example:
        >>> device.getprop("ro.build.id")
        'FP2-gms-18.02.0'

        Arguments:
            key: The key of the property to get.
        Returns:
            The value of a system property.

        """
        with self.adb() as _adb:
            return _adb.Shell(f"getprop {key}").strip()

    def force_awake(self, always: bool = True) -> None:
        """Force the device to be awake while connected.

        Propagate any exception raised by `#adb()`.

        Arguments:
            always: Whether the device should be forced awake or let it handle
                its awake/sleep behaviour.

        """
        with self.adb() as _adb:
            _adb.Shell(
                "svc power stayon {}".format("true" if always else "false")
            )

    def home(self) -> None:
        """Issue a "home button" command.

        Propagate any exception raised by `#adb()`.

        """
        with self.adb() as _adb:
            _adb.Shell("input keyevent KEYCODE_HOME")

    def unlock(self) -> None:
        """Wake-up the device and unlock it.

        The lock-screen protection has to be set to "None" or to "Swipe".

        Propagate any exception raised by `#adb()`.

        TODO: Catch cases when lock-screen protection is not supported.

        """
        with self.adb() as _adb:
            _adb.Shell("input keyevent KEYCODE_MENU")
            if self.sdk >= 24:
                # Android 7 requires double the dose
                _adb.Shell("input keyevent KEYCODE_MENU")
            time.sleep(1)

    def has_app(self, app: App, check_enabled: bool = True) -> bool:
        """Check that an app is installed.

        Propagate any exception raised by `#adb()`.

        Arguments:
            app: The app to check the installation of.
            check_enabled (optional): Whether to validate if the app is enabled
                or not.
        Returns:
            True if the app is present (and if applicable, enabled), False
            otherwise.

        """
        if self.sdk >= 24:
            command = ["cmd", "package", "list", "packages"]
        else:
            command = ["pm", "list", "packages"]
        if check_enabled:
            command.append("-e")
        command.append(app.package)
        with self.adb() as _adb:
            # Look for packages that begin with the app package
            package = _adb.Shell(" ".join(command))
            # Look for an exact match
            return "package:{}".format(app.package) in package.splitlines()

    def install_app(self, app: App, grant_perms: bool = True) -> None:
        """Install an app.

        Propagate any exception raised by `#adb()`.

        Arguments:
            app: The app to install.
            grant_perms: Grant all permissions defined in the app manifest
            (Marshmallow onwards).
        Raises:
            AdbCommandError: If the command failed (i.e. return code < 0).

        """
        with self.adb() as _adb:
            _adb.Install(str(self._apk), grant_permissions=grant_perms)
            if not self.has_app(app, True):
                raise AppInstallationError(
                    self, app, "The app is not installed or enabled"
                )

    def uninstall_app(self, app: App) -> None:
        """Uninstall an app.

        Propagate any exception raised by `#adb()`.

        Arguments:
            app: The app to uninstall.
        Raises:
            AdbCommandError: If the command failed (i.e. return code < 0).

        """
        with self.adb() as _adb:
            _adb.Uninstall(app.package)
            if self.has_app(app, False):
                raise AppInstallationError(
                    self, app, "The app is still installed"
                )
